package com.sf.vertx.init;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.ApplicationArguments;
import org.springframework.boot.ApplicationRunner;
import org.springframework.core.annotation.Order;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Component;

import com.sf.vertx.api.pojo.VertxConfig;
import com.sf.vertx.constans.RedisKeyConfig;
import com.sf.vertx.handle.ApiRateLimitHandler;
import com.sf.vertx.handle.AppConfigHandler;
import com.sf.vertx.handle.AppRateLimitHandler;
import com.sf.vertx.handle.BodyHandler;
import com.sf.vertx.handle.ParameterCheckHandler;
import com.sf.vertx.handle.ProxyHandler;
import com.sf.vertx.handle.RestfulFailureHandler;
import com.sf.vertx.utils.ProxyTool;

import io.vertx.core.Future;
import io.vertx.core.Vertx;
import io.vertx.core.http.HttpClient;
import io.vertx.core.http.HttpServer;
import io.vertx.core.http.HttpServerOptions;
import io.vertx.ext.web.Router;
import io.vertx.ext.web.client.WebClient;
import io.vertx.httpproxy.HttpProxy;
import io.vertx.httpproxy.ProxyContext;
import io.vertx.httpproxy.ProxyInterceptor;
import io.vertx.httpproxy.ProxyResponse;
import lombok.extern.slf4j.Slf4j;

/***
 * 动态构建vertx服务
 * 
 * @author xy
 *
 */
@Slf4j
@Order(value = 10)
@Component
public class DynamicBuildServer implements ApplicationRunner {
	// TODO 后面可以和app挂钩
	@Value("${server.vertx.server.default.port}")
	private Integer serverDefaultPort;
	@Autowired
	private RedisTemplate<String, String> redisTemplate;

	@Autowired
	private RedisKeyConfig redisKeyConfig;

	@Override
	public void run(ApplicationArguments args) throws Exception {
		// 初始化redis key
		redisKeyConfig.init();
		// 从redis同步vertx配置
		AppConfigHandler.initVertxConfig(redisTemplate);

		// 加载vertx、应用配置
		appStartLoadData();
	}

	/***
	 * 应用启动, 从redis读取配置,初始化vertx服务
	 * 
	 * @throws Exception
	 */
	private void appStartLoadData() throws Exception {
		Vertx vertx = AppConfigHandler.createVertx();
		// 从redis同步app配置
		AppConfigHandler.initAllAppConfig(redisTemplate);

		VertxConfig vertxConfig = AppConfigHandler.getVertxConfig();
		// 创建HTTP监听
		// 所有ip都能访问
		HttpServerOptions httpServerOptions = new HttpServerOptions().setHost("0.0.0.0");
		HttpServer server = vertx.createHttpServer(httpServerOptions);
		Router mainHttpRouter = Router.router(vertx);
		Integer serverPort = vertxConfig.getPort() == null ? serverDefaultPort : vertxConfig.getPort();
		log.info("serverPort:{}", serverPort);
		server.requestHandler(mainHttpRouter).listen(serverPort, h -> {
			if (h.succeeded()) {
				log.info("HTTP端口监听成功:{}", serverPort);
			} else {
				log.error("HTTP端口监听失败:{}", serverPort);
			}
		});

//		HttpClientOptions clientOptions = new HttpClientOptions();
//		clientOptions.setMaxPoolSize(20); // 最大连接池大小
//		clientOptions.setConnectTimeout(2000); // 连接超时 毫秒
//		clientOptions.setHttp2KeepAliveTimeout(1);
//		clientOptions.setIdleTimeout(1000); // 连接空闲超时 毫秒
//		HttpClient proxyClient = VERTX.createHttpClient(clientOptions);
		HttpClient proxyClient = vertx.createHttpClient();
		HttpProxy proxy = HttpProxy.reverseProxy(proxyClient);
		proxy.originSelector(request -> Future.succeededFuture(ProxyTool.resolveOriginAddress(request)));
		proxy.addInterceptor(new ProxyInterceptor() {
			@Override
			public Future<ProxyResponse> handleProxyRequest(ProxyContext context) {
//				if(StringUtils.equals(sacAppHeaderKey, "dsafdsfadafhappC")) {
//					// 会跳转到 RestfulFailureHandlerImpl
//					throw new HttpException(10003);
//				}
				return context.sendRequest();
			}

			@Override
			public Future<Void> handleProxyResponse(ProxyContext context) {
				// 调试代码,获取reponse body
//				Filter filter = new Filter();
//				ProxyResponse proxyResponse = context.response();
//				Body body = proxyResponse.getBody();
//				proxyResponse.setBody(Body.body(filter.init(context.request().getURI(), body.stream(), false)));
				// 继续拦截链
				return context.sendResponse();
			}
		});
		WebClient mainWebClient = WebClient.create(vertx);

		String rateLimitModel = vertxConfig.getRateLimitModel();
		rateLimitModel = "local";
		mainHttpRouter.route().handler(ParameterCheckHandler.create())
				.handler(AppRateLimitHandler.create(rateLimitModel))
				.handler(ApiRateLimitHandler.create(rateLimitModel))
				.handler(BodyHandler.create()).handler(ProxyHandler.create(mainWebClient, proxy))
				.failureHandler(RestfulFailureHandler.create());
//		mainHttpRouter.route().handler(ProxyHandler.create(mainWebClient, proxy));
	}

}
